{} Unit History {};

(**) INTERFACE (**)

Uses Objects;

type
  PStackMemory = ^TStackMemory;
  TStackMemory = object(TObject)
    RecSize,
    Count,
    Limit,
    Origin  : word;
    Buffer  : pointer;
    constructor Init(aSize, aLimit : word);
    constructor Load(var S : TStream);
    procedure Store(var S : TStream);
    procedure Write(var Item; Pos : word);
    procedure Read(var Item; Pos : word);
    function Push(var Item) : boolean; virtual;
    function Pop(var Item) : boolean; virtual;
    function UnPop(var Item) : boolean; virtual;
    function ReadAll(var Buf) : word;
    procedure Reset; virtual;
    destructor Done; virtual;
  end;

  PRingStack = ^TRingStack;
  TRingStack = object(TStackMemory)
    Filled : boolean;
    constructor Load(var S : TStream);
    procedure Store(var S : TStream);
    function Push(var Item) : boolean; virtual;
    function Pop(var Item) : boolean; virtual;
    function UnPop(var Item) : boolean; virtual;
    function NextPos : word;
    function PrevPos : word;
    function ValidPop : boolean;
    function ValidUnPop : boolean;
    procedure Reset; virtual;
  end;

  PQueue = ^TQueue;
  TQueue = object(TRingStack)
    function Push(var Item) : boolean; virtual;
    function Pop(var Item) : boolean; virtual;
    function PushBack(var Item) : boolean;
    function CurrentItem : pointer;
    function RemoveItem : boolean;
    function GetItem(Index : word) : pointer;
    function IsFull : boolean;
    function GetSize : word;
  end;

(**) IMPLEMENTATION (**)

{*** TStackMemory ***}

constructor TStackMemory.Init(aSize, aLimit : word);
  var BufSize : longint;
  begin
    BufSize := aSize*aLimit;
    if (BufSize > 65520) or (MaxAvail < BufSize) then Fail;
    inherited Init;
    RecSize := aSize;
    Limit := aLimit;
    GetMem( Buffer, BufSize );
  end;

constructor TStackMemory.Load(var S : TStream);
  var BufSize : word;
  begin
    S.Read(RecSize, 4*SizeOf(word));
    BufSize := RecSize * Limit;
    if MaxAvail < BufSize then Fail;
    GetMem( Buffer, BufSize );
    S.Read( Buffer^, BufSize );
  end;

procedure TStackMemory.Store(var S : TStream);
  begin
    S.Write(RecSize, 4*SizeOf(word));
    S.Write(Buffer^, RecSize * Limit);
  end;

procedure TStackMemory.Write(var Item; Pos : word); assembler;
  asm
    les   di, Self
    mov   bx, ds
    lds   si, Item
    mov   cx, es:[di].TStackMemory.RecSize
    mov   ax, cx
    mul   Pos                               { ax = Pos * RecSize }
    les   di, es:[di].TStackMemory.Buffer
    add   di, ax                            { inc( Buffer.Offs, ax }
    cld
    rep   movsb                             { copy Item to Buffer^ }
    mov   ds, bx
  end;

procedure TStackMemory.Read(var Item; Pos : word); assembler;
  asm
    les   di, Self
    mov   bx, ds
    mov   cx, es:[di].TStackMemory.RecSize
    mov   ax, cx
    mul   Pos                               { ax = Pos * RecSize }
    lds   si, es:[di].TStackMemory.Buffer
    add   si, ax                            { inc( Buffer.Offs, ax }
    les   di, Item
    cld
    rep   movsb                             { copy Item from Buffer^ }
    mov   ds, bx
  end;

function TStackMemory.Push(var Item) : boolean;
  begin
    if Count <> Limit then
      begin
        Write(Item, Count);
        inc(Count);
        Origin := Count;
        Push := true;
      end
    else Push := false;
  end;

function TStackMemory.Pop(var Item) : boolean;
  begin
    if Count <> 0 then
      begin
        dec(Count);
        Read(Item, Count);
        Pop := true;
      end
    else Pop := false;
  end;

function TStackMemory.UnPop(var Item) : boolean;
  begin
    UnPop := false;
    if Count <> Origin then
      if Count < Limit-1 then
        begin
          inc(Count);
          Read(Item, Count);
          UnPop := true;
        end;
  end;

function TStackMemory.ReadAll(var Buf) : word; assembler;
  var Num : word;
  asm
    les   di, Self
    mov   bx, ds

    test  es:[di].TQueue.Filled,  $FF
    jnz   @1

    mov   ax, es:[di].TStackMemory.Count
    sub   ax, es:[di].TStackMemory.Origin
    mov   Num, ax
    mul   es:[di].TStackMemory.RecSize
    mov   cx, ax

    mov   ax, es:[di].TStackMemory.RecSize
    mul   es:[di].TStackMemory.Origin
    lds   si, es:[di].TStackMemory.Buffer
    add   si, ax

    les   di, Buf
    cld
    rep   movsb
    jmp   @E

@1:
    mov   ax, es:[di].TQueue.Limit
    sub   ax, es:[di].TQueue.Origin
    mov   Num, ax
    mul   es:[di].TStackMemory.RecSize
    mov   cx, ax

    mov   ax, es:[di].TStackMemory.RecSize
    mul   es:[di].TStackMemory.Origin
    lds   si, es:[di].TStackMemory.Buffer
    add   si, ax

    les   di, Buf
    cld
    rep   movsb

    lds   si, Self
    mov   ax, [si].TQueue.Count
    add   Num, ax
    mul   [si].TStackMemory.RecSize
    mov   cx, ax
    lds   si, [si].TStackMemory.Buffer
    rep   movsb

@E: mov   ax, Num
    mov   ds, bx
  end;

procedure TStackMemory.Reset;
  begin
    Count := 0;
    Origin := 0;
  end;

destructor TStackMemory.Done;
  begin
    FreeMem( Buffer, RecSize * Limit );
    inherited Done;
  end;


{*** TRingStack ***}

constructor TRingStack.Load(var S : TStream);
  begin
    if inherited Load(S) then S.Read(Filled, SizeOf(Filled))
    else Fail;
  end;

procedure TRingStack.Store(var S : TStream);
  begin
    inherited Store(S);
    S.Write(Filled, SizeOf(Filled));
  end;

function TRingStack.Push(var Item) : boolean;
  begin
    Write(Item, Count);
    Count := NextPos;
    if Count = 0 then Filled := true;
    Origin := Count;
    Push := true;
  end;

function TRingStack.Pop(var Item) : boolean;
  begin
    Pop := false;
    if ValidPop then
      begin
        if Origin = Count then
          begin
            Write(Item, Count);
            Origin := NextPos;
            if Origin = 0 then Filled := true;
          end;
        Count := PrevPos;
        Read(Item, Count);
        Pop := true;
      end
  end;

function TRingStack.UnPop(var Item) : boolean;
  begin
    if ValidUnPop then
      begin
        Count := NextPos;
        Read(Item, Count);
        UnPop := true;
      end
    else UnPop := false;
  end;

procedure TRingStack.Reset;
  begin
    inherited Reset;
    Filled := false;
  end;

function TRingStack.NextPos : word; assembler;
  asm
    les   di, Self
    mov   ax, es:[di].TStackMemory.Count
    inc   ax
    cmp   ax, es:[di].TStackMemory.Limit
    jb    @Exit
    xor   ax, ax
  @Exit:
  end;

function TRingStack.PrevPos : word; assembler;
  asm
    les   di, Self
    mov   ax, es:[di].TStackMemory.Count
    or    ax, ax
    jnz   @Exit
    mov   ax, es:[di].TStackMemory.Limit
  @Exit:
    dec   ax
  end;

function TRingStack.ValidPop : boolean;
  begin
    ValidPop := not ((Count = 0) and not Filled) and (PrevPos <> Origin);
  end;

function TRingStack.ValidUnPop : boolean;
  begin
    ValidUnPop := (Count <> Origin) and (NextPos <> Origin);
  end;


{*** TQueue ***}

function TQueue.Push(var Item) : boolean;
  var F : boolean;
  begin
    Write(Item, Count);
    F := isFull;
    Count := NextPos;
    if Count = 0 then Filled := true;
    if F then Origin := Count;
    Push := not F;
  end;

function TQueue.Pop(var Item) : boolean;
  begin
    Pop := false;
    if GetSize <> 0 then
      begin
        Read(Item, Origin);
        inc(Origin);
        if Origin = Limit then
          begin
            Origin := 0;
            Filled := false;
          end;
        Pop := true;
      end
  end;

function TQueue.PushBack(var Item) : boolean;
  var F : boolean;
  begin
    F := isFull;
    if Origin = 0 then
      begin
        Origin := Limit - 1;
        if not F then Filled := false;
      end
    else dec(Origin);
    Write(Item, Origin);
    if F then Count := Origin;
    PushBack := not f;
  end;

function TQueue.CurrentItem : pointer; assembler;
  asm
    les   di, Self
    mov   ax, es:[di].TStackMemory.RecSize
    mul   es:[di].TStackMemory.Origin       { ax = Origin * RecSize }
    add   ax, word ptr es:[di].TStackMemory.Buffer   { inc( Buffer.Offs, ax }
    mov   dx, word ptr es:[di].TStackMemory.Buffer+2
  end;

function TQueue.GetItem(Index : word) : pointer; assembler;
  asm
    les   di, Self
    mov   ax, es:[di].TStackMemory.Origin
    add   ax, Index
    cmp   ax, es:[di].TStackMemory.Limit
    jb    @MakePtr
    sub   ax, es:[di].TStackMemory.Limit
@MakePtr:
    mul   es:[di].TStackMemory.RecSize      { ax = (Origin+Index) * RecSize }
    add   ax, word ptr es:[di].TStackMemory.Buffer   { inc( Buffer.Offs, ax }
    mov   dx, word ptr es:[di].TStackMemory.Buffer+2
  end;

function TQueue.RemoveItem : boolean;
  begin
    RemoveItem := false;
    if Origin <> Count then
      begin
        inc(Origin);
        if Origin = Limit then
          begin
            Origin := 0;
            Filled := false;
          end;
        if Origin <> Count then RemoveItem := true;
      end;
  end;

function TQueue.IsFull : boolean;
  begin
    IsFull := (Count = Origin) and Filled;
  end;

function TQueue.GetSize : word; assembler;
  asm
    les   di, Self
    test  es:[di].TQueue.Filled,  $FF
    jnz   @1
    mov   ax, es:[di].TQueue.Count
    sub   ax, es:[di].TQueue.Origin
    jmp   @exit
@1:
    mov   ax, es:[di].TQueue.Limit
    sub   ax, es:[di].TQueue.Origin
    add   ax, es:[di].TQueue.Count
@exit:
  end;


end.
